"""sonusai audiofe

usage: audiofe [-hvds] [--version] [-i INPUT] [-l LENGTH] [-m MODEL] [-a ASR] [-w WMODEL]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d, --debug                     Write debug data to H5 file.
    -s, --show                      Display a list of available audio inputs.
    -i INPUT, --input INPUT         Input audio.
    -l LENGTH, --length LENGTH      Length of audio in seconds. [default: -1].
    -m MODEL, --model MODEL         ONNX model.
    -a ASR, --asr ASR               ASR method to use.
    -w WMODEL, --whisper WMODEL     Model used in whisper, aixplain_whisper and faster_whisper methods. [default: tiny].

Aaware SonusAI Audio Front End.

Capture LENGTH seconds of audio from INPUT. If LENGTH is < 0, then capture until key is pressed. If INPUT is a valid
audio file name, then use the audio data from the specified file. In this case, if LENGTH is < 0, process entire file;
otherwise, process min(length(INPUT), LENGTH) seconds of audio from INPUT. Audio is saved to
audiofe_capture_<TIMESTAMP>.wav.

If a model is specified, run prediction on audio data from this model. Then compute the inverse transform of the
prediction result and save to audiofe_predict_<TIMESTAMP>.wav.

Also, if a model is specified, save plots of the capture data (time-domain signal and feature) to
audiofe_capture_<TIMESTAMP>.png and predict data (time-domain signal and feature) to
audiofe_predict_<TIMESTAMP>.png.

If an ASR is specified, run ASR on the captured audio and print the results. In addition, if a model was also specified,
run ASR on the predict audio and print the results.  Examples: faster_whisper, google,

If the debug option is enabled, write capture audio, feature, reconstruct audio, predict, and predict audio to
audiofe_<TIMESTAMP>.h5.

"""
import signal

import numpy as np

from sonusai.mixture import AudioT


def signal_handler(_sig, _frame):
    import sys

    from sonusai import logger

    logger.info('Canceled due to keyboard interrupt')
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    length = float(args['--length'])
    input_name = args['--input']
    model_name = args['--model']
    asr_name = args['--asr']
    whisper_name = args['--whisper']
    debug = args['--debug']
    show = args['--show']

    from os.path import exists

    import h5py
    import pyaudio

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import get_audio_from_feature
    from sonusai.mixture import get_feature_from_audio
    from sonusai.utils import calc_asr
    from sonusai.utils import create_timestamp
    from sonusai.utils import get_input_devices
    from sonusai.utils import load_ort_session
    from sonusai.utils import write_audio

    ts = create_timestamp()
    capture_name = f'audiofe_capture_{ts}'
    capture_wav = capture_name + '.wav'
    capture_png = capture_name + '.png'
    predict_name = f'audiofe_predict_{ts}'
    predict_wav = predict_name + '.wav'
    predict_png = predict_name + '.png'
    h5_name = f'audiofe_{ts}.h5'

    # Setup logging file
    create_file_handler('audiofe.log')
    update_console_handler(verbose)
    initial_log_messages('audiofe')

    if show:
        logger.info('List of available audio inputs:')
        logger.info('')
        p = pyaudio.PyAudio()
        for name in get_input_devices(p):
            logger.info(f'{name}')
        logger.info('')
        p.terminate()
        return

    if input_name is not None and exists(input_name):
        capture_audio = get_frames_from_file(input_name, length)
    else:
        try:
            capture_audio = get_frames_from_device(input_name, length)
        except ValueError as e:
            logger.exception(e)
            return
        # Only write if capture from device, not for file input
        write_audio(capture_wav, capture_audio, SAMPLE_RATE)
        logger.info('')
        logger.info(f'Wrote capture audio with shape {capture_audio.shape} to {capture_wav}')

    if debug:
        with h5py.File(h5_name, 'a') as f:
            if 'capture_audio' in f:
                del f['capture_audio']
            f.create_dataset('capture_audio', data=capture_audio)
        logger.info(f'Wrote capture feature data with shape {capture_audio.shape} to {h5_name}')

    if asr_name is not None:
        logger.info(f'Running ASR on captured audio with {asr_name} ...')
        capture_asr = calc_asr(capture_audio, engine=asr_name, whisper_model_name=whisper_name).text
        logger.info(f'Capture audio ASR: {capture_asr}')

    if model_name is not None:
        session, options, model_root, hparams, sess_inputs, sess_outputs = load_ort_session(model_name)
        if hparams is None:
            logger.error(f'Error: ONNX model does not have required SonusAI hyperparameters, cannot proceed.')
            raise SystemExit(1)
        feature_mode = hparams['feature']
        in0name = sess_inputs[0].name
        in0type = sess_inputs[0].type
        out_names = [n.name for n in session.get_outputs()]

        # frames x stride x feat_params
        feature = get_feature_from_audio(audio=capture_audio, feature_mode=feature_mode)
        save_figure(capture_png, capture_audio, feature)
        logger.info(f'Wrote capture plots to {capture_png}')

        if debug:
            with h5py.File(h5_name, 'a') as f:
                if 'feature' in f:
                    del f['feature']
                f.create_dataset('feature', data=feature)
            logger.info(f'Wrote feature with shape {feature.shape} to {h5_name}')

        if in0type.find('float16') != -1:
            logger.info(f'Detected input of float16, converting all feature inputs to that type.')
            feature = np.float16(feature)  # type: ignore

        # Run inference, ort session wants batch x timesteps x feat_params, outputs numpy BxTxFP or BxFP
        # Note full reshape not needed here since we assume speech enhancement type model, so a transpose suffices
        predict = np.transpose(session.run(out_names, {in0name: np.transpose(feature, (1, 0, 2))})[0], (1, 0, 2))

        if debug:
            with h5py.File(h5_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict)
            logger.info(f'Wrote predict with shape {predict.shape} to {h5_name}')

        predict_audio = get_audio_from_feature(feature=predict, feature_mode=feature_mode)
        write_audio(predict_wav, predict_audio, SAMPLE_RATE)
        logger.info(f'Wrote predict audio with shape {predict_audio.shape} to {predict_wav}')
        if debug:
            with h5py.File(h5_name, 'a') as f:
                if 'predict_audio' in f:
                    del f['predict_audio']
                f.create_dataset('predict_audio', data=predict_audio)
            logger.info(f'Wrote predict audio with shape {predict_audio.shape} to {h5_name}')

        save_figure(predict_png, predict_audio, predict)
        logger.info(f'Wrote predict plots to {predict_png}')

        if asr_name is not None:
            logger.info(f'Running ASR on model-enhanced audio with {asr_name} ...')
            predict_asr = calc_asr(predict_audio, engine=asr_name, whisper_model_name=whisper_name).text
            logger.info(f'Predict audio ASR: {predict_asr}')


def get_frames_from_device(input_name: str | None, length: float, chunk: int = 1024) -> AudioT:
    from select import select
    from sys import stdin

    import pyaudio

    from sonusai import logger
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import get_input_device_index_by_name
    from sonusai.utils import get_input_devices

    p = pyaudio.PyAudio()

    input_devices = get_input_devices(p)
    if not input_devices:
        raise ValueError('No input audio devices found')

    if input_name is None:
        input_name = input_devices[0]

    try:
        device_index = get_input_device_index_by_name(p, input_name)
    except ValueError:
        msg = f'Could not find {input_name}\n'
        msg += f'Available devices:\n'
        for input_device in input_devices:
            msg += f'  {input_device}\n'
        raise ValueError(msg)

    logger.info(f'Capturing from {p.get_device_info_by_index(device_index).get("name")}')
    stream = p.open(format=pyaudio.paFloat32,
                    channels=CHANNEL_COUNT,
                    rate=SAMPLE_RATE,
                    input=True,
                    input_device_index=device_index)
    stream.start_stream()

    print()
    print('+---------------------------------+')
    print('| Press Enter to stop             |')
    print('+---------------------------------+')
    print()

    elapsed = 0.0
    seconds_per_chunk = float(chunk) / float(SAMPLE_RATE)
    raw_frames = []
    while elapsed < length or length == -1:
        raw_frames.append(stream.read(num_frames=chunk, exception_on_overflow=False))
        elapsed += seconds_per_chunk
        if select([stdin, ], [], [], 0)[0]:
            stdin.read(1)
            length = elapsed

    stream.stop_stream()
    stream.close()
    p.terminate()
    frames = np.frombuffer(b''.join(raw_frames), dtype=np.float32)
    return frames


def get_frames_from_file(input_name: str, length: float) -> AudioT:
    from sonusai import logger
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import read_audio

    logger.info(f'Capturing from {input_name}')
    frames = read_audio(input_name)
    if length != -1:
        num_frames = int(length * SAMPLE_RATE)
        if len(frames) > num_frames:
            frames = frames[:num_frames]
    return frames


def save_figure(name: str, audio: np.ndarray, feature: np.ndarray) -> None:
    import matplotlib.pyplot as plt
    from scipy.interpolate import CubicSpline

    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import unstack_complex

    spectrum = 20 * np.log(np.abs(np.squeeze(unstack_complex(feature)).transpose()))
    frames = spectrum.shape[1]
    samples = (len(audio) // frames) * frames
    length_in_s = samples / SAMPLE_RATE
    interp = samples // frames

    ts = np.arange(0.0, length_in_s, interp / SAMPLE_RATE)
    t = np.arange(0.0, length_in_s, 1 / SAMPLE_RATE)

    spectrum = CubicSpline(ts, spectrum, axis=-1)(t)

    fig, (ax1, ax2) = plt.subplots(nrows=2)
    ax1.set_title(name)
    ax1.plot(t, audio[:samples])
    ax1.set_ylabel('Signal')
    ax1.set_xlim(0, length_in_s)
    ax1.set_ylim(-1, 1)

    ax2.imshow(spectrum, origin='lower', aspect='auto')
    ax2.set_xticks([])
    ax2.set_ylabel('Feature')

    plt.savefig(name, dpi=300)


if __name__ == '__main__':
    main()



import pyaudio


def get_input_device_index_by_name(p: pyaudio.PyAudio, name: str = None) -> int:
    info = p.get_host_api_info_by_index(0)
    device_count = info.get('deviceCount')
    for i in range(0, device_count):
        device_info = p.get_device_info_by_host_api_device_index(0, i)
        if name is None:
            device_name = None
        else:
            device_name = device_info.get('name')
        if name == device_name and device_info.get('maxInputChannels') > 0:
            return i

    raise ValueError(f'Could not find {name}')


def get_input_devices(p: pyaudio.PyAudio) -> list[str]:
    names = []
    info = p.get_host_api_info_by_index(0)
    device_count = info.get('deviceCount')
    for i in range(0, device_count):
        device_info = p.get_device_info_by_host_api_device_index(0, i)
        device_name = device_info.get('name')
        if device_info.get('maxInputChannels') > 0:
            names.append(device_name)

    return names


def get_default_input_device(p: pyaudio.PyAudio) -> str:
    info = p.get_host_api_info_by_index(0)
    device_count = info.get('deviceCount')
    for i in range(0, device_count):
        device_info = p.get_device_info_by_host_api_device_index(0, i)
        device_name = device_info.get('name')
        if device_info.get('maxInputChannels') > 0:
            return device_name

    raise ValueError('No input audio devices found')



from typing import Optional

from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Feature


def get_feature_from_audio(audio: AudioT,
                           feature_mode: str,
                           num_classes: Optional[int] = 1,
                           truth_mutex: Optional[bool] = False) -> Feature:
    """Apply forward transform and generate feature data from audio data

    :param audio: Time domain audio data [samples]
    :param feature_mode: Feature mode
    :param num_classes: Number of classes
    :param truth_mutex: Whether to calculate 'other' label
    :return: Feature data [frames, strides, feature_parameters]
    """
    import numpy as np
    from pyaaware import FeatureGenerator

    from .augmentation import pad_audio_to_frame
    from .datatypes import TransformConfig
    from .helpers import forward_transform

    fg = FeatureGenerator(feature_mode=feature_mode,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)

    audio_f = forward_transform(audio=audio,
                                config=TransformConfig(N=fg.ftransform_N,
                                                       R=fg.ftransform_R,
                                                       bin_start=fg.bin_start,
                                                       bin_end=fg.bin_end,
                                                       ttype=fg.ftransform_ttype))

    transform_frames = audio_f.shape[0]
    feature_frames = transform_frames // (fg.decimation * fg.step)
    feature = np.empty((feature_frames, fg.stride, fg.feature_parameters), dtype=np.float32)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        fg.execute(audio_f[transform_frame])

        if fg.eof():
            feature[feature_frame] = fg.feature()
            feature_frame += 1

    return feature


def get_audio_from_feature(feature: Feature,
                           feature_mode: str,
                           num_classes: Optional[int] = 1,
                           truth_mutex: Optional[bool] = False) -> AudioT:
    """Apply inverse transform to feature data to generate audio data

    :param feature: Feature data [frames, stride=1, feature_parameters]
    :param feature_mode: Feature mode
    :param num_classes: Number of classes
    :param truth_mutex: Whether to calculate 'other' label
    :return: Audio data [samples]
    """
    import numpy as np

    from pyaaware import FeatureGenerator

    from sonusai import SonusAIError
    from .datatypes import TransformConfig
    from .helpers import inverse_transform
    from sonusai.utils.stacked_complex import unstack_complex
    from sonusai.utils.compress import power_uncompress

    if feature.ndim != 3:
        raise SonusAIError('feature must have 3 dimensions: [frames, stride=1, feature_parameters]')

    if feature.shape[1] != 1:
        raise SonusAIError('Strided feature data is not supported for audio extraction; stride must be 1.')

    fg = FeatureGenerator(feature_mode=feature_mode,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)

    feature_complex = unstack_complex(feature.squeeze())
    if feature_mode[0:1] == 'h':
        feature_complex = power_uncompress(feature_complex)
    return np.squeeze(inverse_transform(transform=feature_complex,
                                        config=TransformConfig(N=fg.itransform_N,
                                                               R=fg.itransform_R,
                                                               bin_start=fg.bin_start,
                                                               bin_end=fg.bin_end,
                                                               ttype=fg.itransform_ttype)))



from dataclasses import dataclass
from typing import Any
from typing import Iterable
from typing import NamedTuple
from typing import Optional
from typing import SupportsIndex
from typing import TypeAlias

import numpy as np
import numpy.typing as npt
from dataclasses_json import DataClassJsonMixin
from praatio.utilities.constants import Interval

AudioT: TypeAlias = npt.NDArray[np.float32]
AudiosT: TypeAlias = list[AudioT]

ListAudiosT: TypeAlias = list[AudiosT]

Truth: TypeAlias = npt.NDArray[np.float32]
Segsnr: TypeAlias = npt.NDArray[np.float32]

AudioF: TypeAlias = npt.NDArray[np.complex64]
AudiosF: TypeAlias = list[AudioF]

EnergyT: TypeAlias = npt.NDArray[np.float32]
EnergyF: TypeAlias = npt.NDArray[np.float32]

Feature: TypeAlias = npt.NDArray[np.float32]

Predict: TypeAlias = npt.NDArray[np.float32]

# Json type defined to maintain compatibility with DataClassJsonMixin
Json: TypeAlias = dict | list | str | int | float | bool | None


class DataClassSonusAIMixin(DataClassJsonMixin):
    def __str__(self):
        return f'{self.to_dict()}'

    # Override DataClassJsonMixin to remove dictionary keys with values of None
    def to_dict(self, encode_json=False) -> dict[str, Json]:
        def del_none(d):
            if isinstance(d, dict):
                for key, value in list(d.items()):
                    if value is None:
                        del d[key]
                    elif isinstance(value, dict):
                        del_none(value)
                    elif isinstance(value, list):
                        for item in value:
                            del_none(item)
            elif isinstance(d, list):
                for item in d:
                    del_none(item)
            return d

        return del_none(super().to_dict(encode_json))


@dataclass(frozen=True)
class TruthSetting(DataClassSonusAIMixin):
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[list[int]] = None

    def __hash__(self):
        return hash(self.to_json())

    def __eq__(self, other):
        return isinstance(other, TruthSetting) and hash(self) == hash(other)


TruthSettings: TypeAlias = list[TruthSetting]
NumberStr: TypeAlias = float | int | str
OptionalNumberStr: TypeAlias = Optional[NumberStr]
OptionalListNumberStr: TypeAlias = Optional[list[NumberStr]]
EQ: TypeAlias = tuple[float | int, float | int, float | int]


@dataclass
class AugmentationRule(DataClassSonusAIMixin):
    normalize: OptionalNumberStr = None
    pitch: OptionalNumberStr = None
    tempo: OptionalNumberStr = None
    gain: OptionalNumberStr = None
    eq1: OptionalListNumberStr = None
    eq2: OptionalListNumberStr = None
    eq3: OptionalListNumberStr = None
    lpf: OptionalNumberStr = None
    ir: OptionalNumberStr = None
    mixup: Optional[int] = 1


AugmentationRules: TypeAlias = list[AugmentationRule]


@dataclass
class Augmentation(DataClassSonusAIMixin):
    normalize: Optional[float] = None
    pitch: Optional[float] = None
    tempo: Optional[float] = None
    gain: Optional[float] = None
    eq1: Optional[EQ] = None
    eq2: Optional[EQ] = None
    eq3: Optional[EQ] = None
    lpf: Optional[float] = None
    ir: Optional[int] = None


Augmentations: TypeAlias = list[Augmentation]


@dataclass(frozen=True)
class UniversalSNRGenerator:
    is_random: bool
    _raw_value: float | str

    @property
    def value(self) -> float:
        if self.is_random:
            from .augmentation import evaluate_random_rule

            return float(evaluate_random_rule(str(self._raw_value)))

        return float(self._raw_value)


class UniversalSNR(float):
    def __new__(cls, value: float, is_random: bool = False):
        return float.__new__(cls, value)

    def __init__(self, value: float, is_random: bool = False) -> None:
        float.__init__(value)
        self._is_random = bool(is_random)

    @property
    def is_random(self) -> bool:
        return self._is_random


Speaker: TypeAlias = dict[str, str]


@dataclass
class TargetFile(DataClassSonusAIMixin):
    name: str
    samples: int
    truth_settings: TruthSettings
    class_balancing_augmentation: Optional[AugmentationRule] = None
    level_type: Optional[str] = None
    speaker_id: Optional[int] = None

    @property
    def duration(self) -> float:
        from .constants import SAMPLE_RATE

        return self.samples / SAMPLE_RATE


TargetFiles: TypeAlias = list[TargetFile]


@dataclass
class AugmentedTarget(DataClassSonusAIMixin):
    target_id: int
    target_augmentation_id: int


AugmentedTargets: TypeAlias = list[AugmentedTarget]


@dataclass
class NoiseFile(DataClassSonusAIMixin):
    name: str
    samples: int

    @property
    def duration(self) -> float:
        from .constants import SAMPLE_RATE

        return self.samples / SAMPLE_RATE


NoiseFiles: TypeAlias = list[NoiseFile]
ClassCount: TypeAlias = list[int]

GeneralizedIDs: TypeAlias = str | int | list[int] | range


@dataclass(frozen=True)
class TruthFunctionConfig(DataClassSonusAIMixin):
    feature: str
    mutex: bool
    num_classes: int
    target_gain: float
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[list[int]] = None


@dataclass
class GenMixData:
    targets: Optional[AudiosT] = None
    target: Optional[AudioT] = None
    noise: Optional[AudioT] = None
    mixture: Optional[AudioT] = None
    truth_t: Optional[Truth] = None
    segsnr_t: Optional[Segsnr] = None


@dataclass
class GenFTData:
    feature: Optional[Feature] = None
    truth_f: Optional[Truth] = None
    segsnr: Optional[Segsnr] = None


@dataclass
class ImpulseResponseData:
    name: str
    sample_rate: int
    data: AudioT

    @property
    def length(self) -> int:
        return len(self.data)


ImpulseResponseFiles: TypeAlias = list[str]


@dataclass(frozen=True)
class SpectralMask(DataClassSonusAIMixin):
    f_max_width: int
    f_num: int
    t_max_width: int
    t_num: int
    t_max_percent: int


SpectralMasks: TypeAlias = list[SpectralMask]


@dataclass
class Target(DataClassSonusAIMixin):
    file_id: Optional[int] = None
    augmentation: Optional[Augmentation] = None
    gain: Optional[float] = None


Targets: TypeAlias = list[Target]


@dataclass
class Noise(DataClassSonusAIMixin):
    file_id: Optional[int] = None
    augmentation: Optional[Augmentation] = None
    offset: Optional[int] = None


@dataclass
class Mixture(DataClassSonusAIMixin):
    name: Optional[str] = None
    noise: Optional[Noise] = None
    noise_snr_gain: Optional[float] = None
    samples: Optional[int] = None
    snr: Optional[UniversalSNR] = None
    spectral_mask_id: Optional[int] = None
    spectral_mask_seed: Optional[int] = None
    target_snr_gain: Optional[float] = None
    targets: Optional[Targets] = None

    @property
    def noise_id(self) -> int:
        return self.noise.file_id

    @property
    def target_ids(self) -> list[int]:
        return [target.file_id for target in self.targets]

    @property
    def target_augmentations(self) -> list[Augmentation]:
        return [target.augmentation for target in self.targets]


Mixtures: TypeAlias = list[Mixture]


@dataclass(frozen=True)
class TransformConfig:
    N: int
    R: int
    bin_start: int
    bin_end: int
    ttype: str


@dataclass(frozen=True)
class FeatureGeneratorConfig:
    feature_mode: str
    num_classes: int
    truth_mutex: bool


@dataclass(frozen=True)
class FeatureGeneratorInfo:
    decimation: int
    stride: int
    step: int
    feature_parameters: int
    ft_config: TransformConfig
    eft_config: TransformConfig
    it_config: TransformConfig


ASRConfigs: TypeAlias = dict[str, dict[str, Any]]


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    asr_configs: Optional[ASRConfigs] = None
    class_balancing: Optional[bool] = False
    class_labels: Optional[list[str]] = None
    class_weights_threshold: Optional[list[float]] = None
    feature: Optional[str] = None
    impulse_response_files: Optional[ImpulseResponseFiles] = None
    mixtures: Optional[Mixtures] = None
    noise_mix_mode: Optional[str] = 'exhaustive'
    noise_files: Optional[NoiseFiles] = None
    num_classes: Optional[int] = None
    spectral_masks: Optional[SpectralMasks] = None
    target_files: Optional[TargetFiles] = None
    truth_mutex: Optional[bool] = None
    truth_reduction_function: Optional[str] = None


SpeechMetadata: TypeAlias = str | list[Interval] | None


class SnrFMetrics(NamedTuple):
    avg: Optional[float] = None
    std: Optional[float] = None
    db_avg: Optional[float] = None
    db_std: Optional[float] = None


class SnrFBinMetrics(NamedTuple):
    avg: Optional[np.ndarray] = None
    std: Optional[np.ndarray] = None
    db_avg: Optional[np.ndarray] = None
    db_std: Optional[np.ndarray] = None


class SpeechMetrics(NamedTuple):
    pesq: Optional[float] = None
    csig: Optional[float] = None
    cbak: Optional[float] = None
    covl: Optional[float] = None


class AudioStatsMetrics(NamedTuple):
    dco: Optional[float] = None
    min: Optional[float] = None
    max: Optional[float] = None
    pkdb: Optional[float] = None
    lrms: Optional[float] = None
    pkr: Optional[float] = None
    tr: Optional[float] = None
    cr: Optional[float] = None
    fl: Optional[float] = None
    pkc: Optional[float] = None


@dataclass
class MetricDoc:
    category: str
    name: str
    description: str


class MetricDocs(list[MetricDoc]):
    def __init__(self, __iterable: Iterable[MetricDoc]) -> None:
        super().__init__(item for item in __iterable)

    def __setitem__(self, __key: SupportsIndex, __value: MetricDoc) -> None:  # type: ignore
        super().__setitem__(__key, __value)

    def insert(self, __index: SupportsIndex, __object: MetricDoc) -> None:
        super().insert(__index, __object)

    def append(self, __object: MetricDoc) -> None:
        super().append(__object)

    def extend(self, __iterable: Iterable[MetricDoc]) -> None:
        if isinstance(__iterable, type(self)):
            super().extend(__iterable)
        else:
            super().extend(item for item in __iterable)

    @property
    def pretty(self) -> str:
        max_category_len = ((max([len(item.category) for item in self]) + 9) // 10) * 10
        max_name_len = 2 + ((max([len(item.name) for item in self]) + 1) // 2) * 2
        categories: list[str] = []
        for item in self:
            if item.category not in categories:
                categories.append(item.category)

        result = ''
        for category in categories:
            result += f'{category}\n'
            result += '-' * max_category_len + '\n'
            for item in [sub for sub in self if sub.category == category]:
                result += f'  {item.name:<{max_name_len}}{item.description}\n'
            result += '\n'

        return result

    @property
    def names(self) -> set[str]:
        return set(item.name for item in self)



"""sonusai

usage: sonusai [--version] [--help] <command> [<args>...]

The sonusai commands are:
    <This information is automatically generated.>

Aaware Sound and Voice Machine Learning Framework. See 'sonusai help <command>'
for more information on a specific command.

"""
import signal


def signal_handler(_sig, _frame):
    import sys

    from sonusai import logger

    logger.info('Canceled due to keyboard interrupt')
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


def main() -> None:
    from importlib import import_module
    from pkgutil import iter_modules

    from sonusai import commands_list

    plugins = {}
    plugin_docstrings = []
    for _, name, _ in iter_modules():
        if name.startswith('sonusai_') and not name.startswith('sonusai_asr_'):
            module = import_module(name)
            plugins[name] = {
                'commands': commands_list(module.commands_doc),
                'basedir':  module.BASEDIR,
            }
            plugin_docstrings.append(module.commands_doc)

    from docopt import docopt

    from sonusai import __version__
    from sonusai.utils import add_commands_to_docstring
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(add_commands_to_docstring(__doc__, plugin_docstrings)),
                  version=__version__,
                  options_first=True)

    command = args['<command>']
    argv = args['<args>']

    import sys
    from os.path import join
    from subprocess import call

    import sonusai
    from sonusai import logger

    base_commands = sonusai.commands_list()
    if command == 'help':
        if not argv:
            exit(call(['sonusai', '-h']))
        elif argv[0] in base_commands:
            exit(call(['python', f'{join(sonusai.BASEDIR, argv[0])}.py', '-h']))

        for plugin, data in plugins.items():
            if argv[0] in data['commands']:
                exit(call(['python', f'{join(data["basedir"], argv[0])}.py', '-h']))

        logger.error(f"{argv[0]} is not a SonusAI command. See 'sonusai help'.")
        sys.exit(1)

    if command in base_commands:
        exit(call(['python', f'{join(sonusai.BASEDIR, command)}.py'] + argv))

    for plugin, data in plugins.items():
        if command in data['commands']:
            exit(call(['python', f'{join(data["basedir"], command)}.py'] + argv))

    logger.error(f"{command} is not a SonusAI command. See 'sonusai help'.")
    sys.exit(1)


if __name__ == '__main__':
    main()



import os
from pathlib import Path
from typing import Optional

from .types import TimeAlignedType


def load_text(audio: str | os.PathLike[str]) -> Optional[TimeAlignedType]:
    """Load time-aligned text data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A TimeAlignedType object.
    """
    import string

    from sonusai.mixture import get_sample_rate

    file = Path(audio).with_suffix('.TXT')
    if not os.path.exists(file):
        return None

    sample_rate = get_sample_rate(str(audio))

    with open(file, mode='r', encoding='utf-8') as f:
        line = f.read()

    fields = line.strip().split()
    start = int(fields[0]) / sample_rate
    end = int(fields[1]) / sample_rate
    text = ' '.join(fields[2:]).lower().translate(str.maketrans('', '', string.punctuation))

    return TimeAlignedType(start, end, text)


def load_words(audio: str | os.PathLike[str]) -> Optional[list[TimeAlignedType]]:
    """Load time-aligned word data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A list of TimeAlignedType objects.
    """

    return _load_ta(audio, 'words')


def load_phonemes(audio: str | os.PathLike[str]) -> Optional[list[TimeAlignedType]]:
    """Load time-aligned phonemes data given a TIMIT audio file.

    :param audio: Path to the TIMIT audio file.
    :return: A list of TimeAlignedType objects.
    """

    return _load_ta(audio, 'phonemes')


def _load_ta(audio: str | os.PathLike[str], tier: str) -> Optional[list[TimeAlignedType]]:
    from sonusai.mixture import get_sample_rate

    if tier == 'words':
        file = Path(audio).with_suffix('.WRD')
    elif tier == 'phonemes':
        file = Path(audio).with_suffix('.PHN')
    else:
        raise ValueError(f'Unknown tier: {tier}')

    if not os.path.exists(file):
        return None

    sample_rate = get_sample_rate(str(audio))

    entries: list[TimeAlignedType] = []
    first = True
    with open(file, mode='r', encoding='utf-8') as f:
        for line in f.readlines():
            fields = line.strip().split()
            start = int(fields[0]) / sample_rate
            end = int(fields[1]) / sample_rate
            text = ' '.join(fields[2:])

            if first:
                first = False
            else:
                if start < entries[-1].end:
                    start = entries[-1].end - (entries[- 1].end - start) // 2
                    entries[-1] = TimeAlignedType(text=entries[-1].text, start=entries[-1].start, end=start)

                if end <= start:
                    end = start + 1 / sample_rate

            entries.append(TimeAlignedType(text=text, start=start, end=end))

    return entries


def _years_between(record, born):
    try:
        rec_fields = [int(x) for x in record.split('/')]
        brn_fields = [int(x) for x in born.split('/')]
        return rec_fields[2] - brn_fields[2] - ((rec_fields[1], rec_fields[0]) < (brn_fields[1], brn_fields[0]))
    except ValueError:
        return '??'


def _decode_dialect(d: str) -> str:
    if d in ['DR1', '1']:
        return 'New England'
    if d in ['DR2', '2']:
        return 'Northern'
    if d in ['DR3', '3']:
        return 'North Midland'
    if d in ['DR4', '4']:
        return 'South Midland'
    if d in ['DR5', '5']:
        return 'Southern'
    if d in ['DR6', '6']:
        return 'New York City'
    if d in ['DR7', '7']:
        return 'Western'
    if d in ['DR8', '8']:
        return 'Army Brat'

    raise ValueError(f'Unrecognized dialect: {d}')


def load_speakers(input_dir: Path) -> dict:
    speakers = {}
    with open(input_dir / 'SPKRINFO.TXT') as file:
        for line in file:
            if not line.startswith(';'):
                fields = line.strip().split()
                speaker_id = fields[0]
                gender = fields[1]
                dialect = _decode_dialect(fields[2])
                age = _years_between(fields[4], fields[5])
                speakers[speaker_id] = {'gender': gender, 'dialect': dialect, 'age': age}
    return speakers
